#include <stdio.h>
#include <strings.h>
#include <math.h>

#define PI 3.141592654
#define TPI 6.283185307
#define MAXLEN 256
#define DAY2SEC 86400e0
#define MAXDAT 5000

/***
 NPSD ver1.0 Feb28,1996 K.Hayashida Osaka Univ
      ver1.1 Jun10,1996 K.Hayashida Osaka Univ
        (light curve segment with ndat=0 is ignored)
      ver1.2 Jun13,1996 K.Hayashida Osaka Univ
        (Normalization is done at the end not for each segment)
      ver1.3 Dec26,1996 K.Hayashida Osaka Univ
        (ndat >1 -> ndat >0
*/
void ClearMat(
  double x[],
  int n
)
{ int i;
  for(i=0;i<n;i++) x[i]=0.0e0;
}

void ClearIMat(
  int x[],
  int n
)
{ int i;
  for(i=0;i<n;i++) x[i]=0.0e0;
}

void AveStd(
  double x[],
  int n,
  double *average,
  double *std
)

{ int i;
  double sum,sum2,dn;

  dn = (double) n;
  sum=0; sum2=0;
  for(i=0;i<n;i++) {
    sum = sum + x[i];
    sum2 = sum2 + x[i] * x[i];
  }
  *average = sum / dn;
  *std = sqrt ( ( sum2 - (*average) * dn ) /(dn-1.0) ); 
}

void SquaredMean(
  double x[],
  int n,
  double *sqmean
)

{ int i;
  double sum2,dn;

  dn = (double) n;
  sum2=0;
  for(i=0;i<n;i++) {
    sum2 = sum2 + x[i] * x[i];
  }
  *sqmean = sqrt( (sum2/dn) );
}

/* for ASCA qdp */
int GetLcurveSegment_ASCA(
  FILE *fp,
  int *initflg,
  double *tmjd0,
  double gapsec,
  double tsec[],
  double flux[],
  double ferr[]
)

{  
   int ndat;
   char buff[MAXLEN+1];
   static double tmjd, x, xerr, dtsec;
   double sec, terr, tascasec;

   ndat = 0;
     
   while ( fgets(buff, MAXLEN ,fp) != NULL ) {
      if( ndat == 0 && ( (*initflg) != 1)  ) {
         *tmjd0 = tmjd;
         tsec[ndat] = 0.0; 
         flux[ndat] = x;
         ferr[ndat] = xerr;
       /*  printf(" %d %f %f %f %f\n",ndat,tascasec,tsec[ndat],x,xerr); */
         ndat++;
       }
       /* printf("%s \n", buff); */
       if(buff[0] != 'R' && buff[0] != 'l' && buff[0] != '!') {
         if(sscanf(buff,"%lf %lf %lf %lf",&tascasec,&terr,&x,&xerr) > 0 ) {
            tmjd = tascasec / (DAY2SEC) ; 
            if( (*initflg) == 1) {
              *tmjd0 = tmjd;
              (*initflg) = 0;
	    }
            sec = ( (tmjd - (*tmjd0)) *(DAY2SEC)) ;
	    if (ndat > 0 ) {
              dtsec = sec - tsec[ndat-1];
            } else {
              dtsec = 0.0;
            }
            if ( fabs(dtsec) > gapsec ) {
              /* printf(" gap %d %f %f %f %f\n",ndat,tmjd,sec,x,xerr); */
              return(ndat);
            } else {
              tsec[ndat] = sec; 
              flux[ndat] = x;
              ferr[ndat] = xerr;
          /*  printf(" %d %f %f %f %f\n",ndat,tmjd,tsec[ndat],x,xerr); */
              ndat++;
            }   
          } /* end if scanf */
        } /* end if buff */
    } /* end while */ 
    return(ndat);
}

/* for GINGA   */
int GetLcurveSegment_GINGA(
  FILE *fp,
  int *initflg,
  double *tmjd0,
  double gapsec,
  double tsec[],
  double flux[],
  double ferr[]
)

{  
   int ndat;
   char buff[MAXLEN+1];
   static double tmjd, x, xerr, dtsec;
   double sec, terr;

   ndat = 0;
     
   while ( fgets(buff, MAXLEN ,fp) != NULL ) {
      if( ndat == 0 && ( (*initflg) != 1)  ) {
         *tmjd0 = tmjd;
         tsec[ndat] = 0.0; 
         flux[ndat] = x;
         ferr[ndat] = xerr;
       /*  printf(" %d %f %f %f %f\n",ndat,tmjd,tsec[ndat],x,xerr); */
         ndat++;
       }
       /* printf("%s \n", buff); */
       if(buff[0] != '#' && buff[0] != '!') {
         if(sscanf(buff,"%lf %lf %lf",&tmjd,&x,&xerr) > 0 ) {
            if( (*initflg) == 1) {
              *tmjd0 = tmjd;
              (*initflg) = 0;
	    }
            sec = ( (tmjd - (*tmjd0)) *(DAY2SEC)) ;
	    if (ndat > 0 ) {
              dtsec = sec - tsec[ndat-1];
            } else {
              dtsec = 0.0;
            }
            if ( fabs(dtsec) > gapsec ) {
            /*  printf(" gap %d %f %f %f %f\n",ndat,tmjd,sec,x,xerr); */ 
              return(ndat);
            } else {
              tsec[ndat] = sec; 
              flux[ndat] = x;
              ferr[ndat] = xerr;
   /*         printf(" %d %f %f %f %f\n",ndat,tmjd,tsec[ndat],x,xerr); */
              ndat++;
            }   
          } /* end if scanf */
        } /* end if buff */
    } /* end while */ 
    return(ndat);
}


int AddNPSD(
  double fhz_tmp[],
  double psd_tmp[],
  int nf_tmp,
  double fhz_boundary[],
  int nf,
  double fhz[],
  double psd[],
  double psd2[],
  int nset[]
)

{ int k, k0;

  for(k=0; k < nf_tmp; k++) {
     k0 = 0;
     while ( (fhz_tmp[k] > fhz_boundary[k0+1]) && (k0 < nf ) ) k0++;
      /* fhz_boundary[k0] <= fhz_tmp[k] < fhz_boundary[k0+1] */
     fhz[k0] = ( fhz[k0] * nset[k0] + fhz_tmp[k]) / (nset[k0] + 1.0 );
     psd[k0] = ( psd[k0] * nset[k0] + psd_tmp[k]) / (nset[k0] + 1.0 );
     psd2[k0] = ( psd2[k0] * nset[k0] + psd_tmp[k] * psd_tmp[k]) 
          / (nset[k0] + 1.0 );
     nset[k0]++;
     /* printf(" %d %d %d \n",k,k0,nset[k0]); */
  } /* end of for */   
}

void PutNPSD(
  int nf,
  double fhz[],
  double psd[],
  double psd2[],
  int nset[]
)

{
 int k;
 double perr, dummy;
 char   file0[64];
 FILE   *fp0;

  printf("Output Filename?"); 
  scanf("%s",file0);

  if(NULL == (fp0 = fopen(file0, "w"))){
     printf("\n Cannot Open File : %s \n\n", file0);
     exit(1);
   }
 
 dummy= 0.0;
 for( k=0; k<nf;k++) {
   if(nset[k]>1) {
      perr = sqrt (  ( psd2[k] - psd[k]*psd[k] )  
                         / ((double) (nset[k]-1) )   );
      fprintf(fp0,"%5d %10e %10e %10e %10e %10e %5d\n",
     k,fhz[k], dummy,psd[k], perr,perr,nset[k] );
   } /* end of if */
 } /* end of for */   

}

void PSDtoNPSD(
  int nf,
  double psd[],
  double psd2[],
  double fave,
  double bgdrate
)

{
 int k;

 for( k=0; k<nf;k++) {
   psd[k] = psd[k] /(fave-bgdrate)/(fave-bgdrate);
   psd2[k] = psd2[k] /(fave-bgdrate)/(fave-bgdrate)
    /(fave-bgdrate)/(fave-bgdrate);
 } /* end of for */   

}


int CalcNPSD(
 double tsec[],
 double flux[],
 double ferr[],
 int  ndat,
 double binsec,
 double bgdrate,
 double fhz[],
 double psd[]
)

{
 double tdel, bins, fave, dummy, pois;
 int nf, jf, i;
 double freq, omg, tau, s, c, p1, p2, p3 ,p4 ;

 tdel = tsec[ndat-1] - tsec[0];
 bins = (double) ndat;
 AveStd(flux,ndat,&fave,&dummy); 
 SquaredMean(ferr,ndat,&pois); 
 printf("# ndat %d fave %f  pois %f \n", ndat,fave, pois);
 nf = ndat/2;
 
 for(jf=0;jf<nf;jf++) {
    freq =   ((double)(jf+1)) /tdel; 
    if(tdel*freq > 0.9 ) {
      fhz[jf] = freq ;
      omg = (TPI) * fhz[jf];
      s = 0.0 ;
      c = 0.0 ; 
      for(i=0;i<ndat;i++) {
        s = s + sin( fmod(2.0 * omg * tsec[i],TPI));
        c = c + cos( fmod(2.0 * omg * tsec[i],TPI));
    /*  printf(" freq %f omg %f s %f c %f tsec %f \n",freq,omg,s,c,tsec[i]); */
      } /* end of for i */
      tau = atan2 (s,c) /2.0/omg;
      p1=0.0;p2=0.0;p3=0.0;p4=0.0;
      for(i=0;i<ndat;i++) {
         p1 = p1 + (flux[i] - fave) 
               * cos(fmod(omg * (tsec[i]-tau),TPI));
         p2 = p2 + cos(fmod(omg * (tsec[i]-tau),TPI))
               * cos(fmod(omg * (tsec[i]-tau),TPI));
         p3 = p3 + (flux[i] - fave) 
               * sin(fmod(omg * (tsec[i]-tau),TPI));
         p4 = p4 + sin(fmod(omg * (tsec[i]-tau),TPI))
               * sin(fmod(omg * (tsec[i]-tau),TPI));
    /*    printf(" omg tsec tau i jf = %f %f %f %d %d\n",omg,tsec[i],tau,i,jf);
     */
      } /* end of for i */
    /*  psd[jf] = (p1*p1/p2 + p3*p3/p4)/2.0/ bins * tdel; */ 
     psd[jf] = (p1*p1 + p3*p3)/ bins/bins * tdel;  
  /* if( p4<1e-10 ) printf(" omg p4 tau nf = %f %e %f %d\n",omg,p4,tau,nf);
   */ 
      psd[jf] = psd[jf] - pois*pois/bins*tdel;    
      psd[jf] = psd[jf] / (fave-bgdrate)/(fave-bgdrate); 
     /*  prvintf(" %d %e Hz  NPSD %e \n",jf,freq,psd[jf]); */
    } /* end of if */
  } /* end of for jf */
  return(nf);
}


int CalcPSD(
 double tsec[],
 double flux[],
 double ferr[],
 int  ndat,
 double binsec,
 double fhz[],
 double psd[],
 double *fave_temp
)

{
 double tdel, bins, fave, dummy, pois;
 int nf, jf, i;
 double freq, omg, tau, s, c, p1, p2, p3 ,p4 ;

 tdel = tsec[ndat-1] - tsec[0];
 bins = (double) ndat;
 AveStd(flux,ndat,&fave,&dummy); 
 SquaredMean(ferr,ndat,&pois); 
 printf("# ndat %d fave %f  pois %f \n", ndat,fave, pois);
 nf = ndat/2;
 
 for(jf=0;jf<nf;jf++) {
    freq =   ((double)(jf+1)) /tdel; 
    if(tdel*freq > 0.9 ) {
      fhz[jf] = freq ;
      omg = (TPI) * fhz[jf];
      s = 0.0 ;
      c = 0.0 ; 
      for(i=0;i<ndat;i++) {
        s = s + sin( fmod(2.0 * omg * tsec[i],TPI));
        c = c + cos( fmod(2.0 * omg * tsec[i],TPI));
    /*  printf(" freq %f omg %f s %f c %f tsec %f \n",freq,omg,s,c,tsec[i]); */
      } /* end of for i */
      tau = atan2 (s,c) /2.0/omg;
      p1=0.0;p2=0.0;p3=0.0;p4=0.0;
      for(i=0;i<ndat;i++) {
         p1 = p1 + (flux[i] - fave) 
               * cos(fmod(omg * (tsec[i]-tau),TPI));
         p2 = p2 + cos(fmod(omg * (tsec[i]-tau),TPI))
               * cos(fmod(omg * (tsec[i]-tau),TPI));
         p3 = p3 + (flux[i] - fave) 
               * sin(fmod(omg * (tsec[i]-tau),TPI));
         p4 = p4 + sin(fmod(omg * (tsec[i]-tau),TPI))
               * sin(fmod(omg * (tsec[i]-tau),TPI));
    /*    printf(" omg tsec tau i jf = %f %f %f %d %d\n",omg,tsec[i],tau,i,jf);
     */
      } /* end of for i */
      psd[jf] = (p1*p1 + p3*p3)/ bins/bins * tdel;  
      psd[jf] = psd[jf] - pois*pois/bins*tdel;    
     /*  psd[jf] = psd[jf] / (fave-bgdrate)/(fave-bgdrate); */
    } /* end of if */
  } /* end of for jf */
  *fave_temp = fave;
  return(nf);
}

main(int argc, char *argv[])
{
  char fname[256];
  FILE *fpin;
  int initflg, nf_tmp, ndat, jf, nf;
  double tsec[MAXDAT], flux[MAXDAT], ferr[MAXDAT];
  double binsec, gapsec, tmjd0;
  double fhz_tmp[MAXDAT], psd_tmp[MAXDAT];
  double fhz[MAXDAT], psd[MAXDAT], psd2[MAXDAT];
  double fhz_boundary[MAXDAT], fmin, fmax;
  int nset[MAXDAT];
  double bgdrate, fave, sum_fave, sum_ndat;

  fprintf(stdout,"# Input BGD subtracted Light Curve File Name \n");
  fscanf(stdin,"%s",fname);
  fprintf(stdout,"# binsec, maxgapsec_allowed \n");
  fscanf(stdin,"%lf %lf",&binsec,&gapsec);
  fprintf(stdout,"# fmin(Hz) fmax(Hz) nfbin  \n");
  fscanf(stdin,"%lf %lf %d",&fmin,&fmax,&nf);
  for(jf=0;jf<= nf;jf++) {
    fhz_boundary[jf] = fmin * pow(10.0, 
                   ( ((double)jf)/((double) nf)  * log10(fmax/fmin)) ) ;
    /* printf("%d %f\n",jf, fhz_boundary[jf]); */
  }
  fpin = fopen(fname,"r");
  initflg = 1;
  ClearMat(psd, MAXDAT);
  ClearMat(psd2, MAXDAT);
  ClearIMat(nset, MAXDAT);

  if( argc > 1 && (strcmp(argv[1],"-g") == 0) ) {
       bgdrate = 0.0;
    while ( (ndat = GetLcurveSegment_GINGA(fpin,&initflg,&tmjd0, gapsec,
                  tsec, flux, ferr)
           ) > 0 ) 
     {
       if(ndat>1) {
	 nf_tmp = CalcNPSD(tsec, flux, ferr,ndat,binsec, bgdrate,
        fhz_tmp, psd_tmp);
	 AddNPSD(fhz_tmp, psd_tmp, nf_tmp, fhz_boundary, nf, 
           fhz, psd, psd2, nset );
       }
     } /* end of while */ 
  } else {
   if( argc > 2 && (strcmp(argv[1],"-b") == 0) ) {  
       bgdrate= atof(argv[2]);
       fprintf(stdout,"# bgdrate %f\n",bgdrate); 
     } else {
       bgdrate = 0.0;
     }
    sum_fave =0.0; 
    sum_ndat =0;
    while ( (ndat = GetLcurveSegment_ASCA(fpin,&initflg,&tmjd0, gapsec,
                  tsec, flux, ferr)
           ) > 0 ) 
     {
       if(ndat>1) {
         nf_tmp = CalcPSD(tsec, flux, ferr,ndat,binsec,
               fhz_tmp, psd_tmp, &fave);
         sum_fave = sum_fave + fave * ndat;
         sum_ndat = sum_ndat + ndat*1.0;
         AddNPSD(fhz_tmp, psd_tmp, nf_tmp, fhz_boundary, nf, 
           fhz, psd, psd2, nset );
       }
      } /* end of while */ 
      fave = sum_fave / sum_ndat;
      PSDtoNPSD(nf, psd, psd2, fave, bgdrate);
  }
  PutNPSD(nf, fhz, psd, psd2, nset);
}

